import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;

import util.BinUtil;
import util.GzipUtil;
import util.InByteBuffer;
import util.ProtocolUtil;
import util.SerialUtil;
import util.Str8Buff;
import util.StreamUtils;
import util.ThreadUtil;
import crypt.Cipher;

/**
 * @author Heiner Kcker
 */
public class ProtocolClient
{
  // --- static marker, once per application ---
  
  static volatile short requestId = Short.MIN_VALUE;

  /**
   * Get and increment request id.
   * @return
   */
  private static synchronized short getNextRequestId()
  {
    return requestId++;
  }
  
  static volatile int sessionId;
  
  static byte[] keyByteArr = "r1U2m3pelsTielzchen".getBytes();

  // --- Const ---
  
  private static final int MAX_TRY_COUNT = 1; // 5;
  
  private static final int TRY_RELAX_TIME = 3 * 1000;
  
  private static final int CONNECTION_TIME_OUT = 60 * 1000;
  
  
  // --- Member ---

  private final String hostName;
  
  private final int portNr;
  
  private Socket socket;

  private int tryCount;

  /**
   * Constructor private,
   * Nicht instanziieren
   */
  public ProtocolClient(
      final String pHostName, //--
      final int pPortNr )
  {
    this.hostName = pHostName;
    this.portNr = pPortNr;
  }
  
  /**
   * 
   * @param pOpCode
   * @param pParamObjArr
   * @param pCompressParams
   * @return
   */
  public Object[] sendRequest(
      final short pOpCode,
      final Object[] pParamObjArr ,
      final boolean pCompressParams)
  {
    tryCount = 0;
    while ( tryCount < MAX_TRY_COUNT)
    {
      try
      {
        // --- dieses if/while nur fr non blocked server ---
        long outTime = System.currentTimeMillis() + CONNECTION_TIME_OUT;
        while ( socket == null || socket.isClosed() )
        // --------------------------------------------
        {
          try
          {
            socket = new Socket(hostName, portNr);
          }
          catch (Exception e)
          {
            // TODO: Log
            System.err.println( e.getMessage() );
            e.printStackTrace();
            
            // timeout logic
            if ( System.currentTimeMillis() > outTime )
            {
              continue;
            }
          }
          //socket.setSoTimeout();
        }
        return innerSendRequest( pOpCode, pParamObjArr, pCompressParams);
      }
      catch (IOException e)
      {
        // TODO log output
        System.err.println( e.getMessage() );
        e.printStackTrace();
        
        ThreadUtil.sleepUnhandled( TRY_RELAX_TIME );
        tryCount++;
      }
    }
    return null;
  }
  
  /**
   * 
   * @param pOperationCode
   * @param pParamObjArr
   * @param pCompressParams
   * @return
   * @throws IOException
   */
  private Object[] innerSendRequest(
      final short pOperationCode,
      final Object[] pParamObjArr ,
      final boolean pCompressParams) 
    throws IOException
  {
    /*final*/ Str8Buff reqBuff = new Str8Buff( 1024 );
    
    // request id
    reqBuff.append( BinUtil.shortToByteArr( getNextRequestId() ) );
    
    // session id, TODO byte array zwischenspeichern
    reqBuff.append( BinUtil.intToByteArr( sessionId ) );

    // operation code
    reqBuff.append( BinUtil.shortToByteArr( pOperationCode ) );

    // params
    if ( pParamObjArr == null || pParamObjArr.length < 1 )
    {
      // keine Parameter
      reqBuff.append( 0 );
    }
    else
    {
      if ( pParamObjArr.length > 127 )
      {
        throw new RuntimeException( "too many parameters " + pParamObjArr.length );
      }
      
      if ( pCompressParams )
      {
        // compress params, indicated over highest bit in param count byte
        reqBuff.append( (byte) ( pParamObjArr.length * -1 ) );
      }
      else
      {
        reqBuff.append( (byte) pParamObjArr.length );
      }
      
      final Str8Buff paramBuff = SerialUtil.serializeMulti( pParamObjArr );
      byte[] paramByteArr = paramBuff.bytes();
      if (pCompressParams)
      {
        paramByteArr = GzipUtil.gzip( paramByteArr );
      }
      reqBuff.append( paramByteArr );
    }
    
    // crypt request
    // neue instanz des Cipher, damit mehrere Threads diese methode durchlaufen knnen
    final Cipher cipher = new Cipher( keyByteArr );
    //fr debug off: cipher.cryptIn( reqBuff );

    // get output stream
    final OutputStream outputStream = socket.getOutputStream();

    // send length of request
    outputStream.write( BinUtil.intToByteArr( reqBuff.length() ) );
System.out.print( SerialUtil.toArrList( BinUtil.intToByteArr( reqBuff.length() ) ) );

    
    // send request header and body
    outputStream.write( reqBuff.innerByteArr(), 0 , reqBuff.length() );
    outputStream.flush();
System.out.print( SerialUtil.toArrList( reqBuff.bytes() ) );

System.out.println();    
    
    // free memory
    reqBuff = null;

    // hier ist zeit zum aufrumen, denn die response muss erst mal kommen
    System.gc();
    System.gc();
    System.gc();
    
    // read response
    Str8Buff respBuff = ProtocolUtil.readFromSockForLength(socket);

    // decrypt response
    cipher.reset();
//  fr debug off: cipher.decryptIn( respBuff );
    
    // get return value count
    // vorzeichen beachten
    int retValueCount = respBuff.byteAt( 8 );
    
    // negativ param count value indicates a compressed body
    final boolean isRespCompressed = retValueCount < 0;

    // TODO check req id and session id
    
    // data bytes
    byte[] respByteArr = respBuff.sub( 9 );
    
    // free memory
    respBuff = null;
    
    if ( isRespCompressed )
    {
      retValueCount *= - 1;
      // decompress response
      respByteArr = GzipUtil.gunzip( respByteArr );
    }
    
    final InByteBuffer inByteBuffer = new InByteBuffer( respByteArr );
    
    final Object[] retObjArr = SerialUtil.deserializeMulti( inByteBuffer , retValueCount );
    
    return retObjArr;
  }

  /**
   * close socket
   */
  public void close()
  {
    StreamUtils.close( socket );
    socket = null;
  }
  
  /**
   * @param args
   */
  public static void main(
      String[] args )
  {
    // TODO Auto-generated method stub

  }

}
