package crypt;

import java.util.Arrays;

import util.Str8;
import util.Str8Buff;

/**
 * Hilfsklasse zur Verschlsselung.
 * 
 * @author Heiner Kcker
 * 
 * TODO wenn es nur ein Exemplar pro App gibt auf static umstellen.
 * 
 * TPDO einen Cipher fr RandomAccess (speichern lokal) machen
 */
public class Cipher
{

  /**
   * Offset fr jedes Byte
   */
  static final int BASIC_OFFSET = 3;

  /** Schlsselwort */
  private final byte[] cryptKeyByteArr;

  /** Position im Schlsselwort */
  private int keyPos = 0;

  /** Umschalgspunkt (Maximum) fr Sgezahn-Verschlsselung */
  static final int SAEGEZAHN_MAX = 27;

  /** Position fr Sgezahn verschlsselung */
  private int saegeZahnPos = 0;

  /**
   * Constructor
   */
  public Cipher( //--
      final byte[] pCryptKeyByteArr)
  {
    super();

    if (pCryptKeyByteArr == null
        || pCryptKeyByteArr.length < 1)
    {
      throw new IllegalArgumentException();
    }
    this.cryptKeyByteArr = pCryptKeyByteArr;
  }

  public void reset()
  {
    keyPos = 0;
    saegeZahnPos = 0;
  }
  
  /**
   * Crypt an single byte.
   * 
   * @param pByte
   * @return
   */
  public int cryptOne(
      final int pByte)
  {
    int cryptByte = pByte;

    // BASIC
    cryptByte = ( pByte + BASIC_OFFSET ) % 256;

    // Schlsselwort
    cryptByte = ( cryptByte + this.cryptKeyByteArr[keyPos] ) % 256;

    incrementKeyPos();

    // Sgezahn
    cryptByte = ( cryptByte + saegeZahnPos ) % 256;

    incrementSaegezahn();
    return cryptByte;
  }

  /**
   * decrypt an single byte.
   * 
   * @param pByte
   * @return
   */
  public int decryptOne( //--
      final int pByte)
  {
    int retByte = pByte;
    
    // Sgezahn
    retByte = ( retByte - saegeZahnPos ) % 256;
    if (retByte < 0)
    {
      retByte += 256;
    }

    incrementSaegezahn();

    // Schlsselwort
    retByte = ( retByte - this.cryptKeyByteArr[keyPos] ) % 256;
    if (retByte < 0)
    {
      retByte += 256;
    }

    incrementKeyPos();

    // BASIC
    retByte = ( retByte - BASIC_OFFSET ) % 256;
    if (retByte < 0)
    {
      retByte += 256;
    }

    return retByte;

  }

  /**
   * TODO einbau strategie
   */
  private void incrementKeyPos()
  {
    keyPos++ ;
    if (keyPos >= this.cryptKeyByteArr.length)
    {
      keyPos = 0;
    }
  }
  
  /**
   * TODO einbau strategie
   */
  private void incrementSaegezahn()
  {
    saegeZahnPos++ ;
    if (saegeZahnPos >= SAEGEZAHN_MAX)
    {
      saegeZahnPos = 0;
    }
  }
  
  /**
   * Crypt the specified byte array.
   * 
   * @param pUncryptByteArr
   * @return
   */
  public byte[] crypt(
      final byte[] pUncryptByteArr)
  {
    final byte[] cryptByteArr = new byte[ pUncryptByteArr.length ];
    for (int i = 0; i < pUncryptByteArr.length; i++ )
    {
      byte b = pUncryptByteArr[i];
      cryptByteArr[i] = (byte) cryptOne( Str8.byteToInt( b ) );
    }
    return cryptByteArr;
  }
  
  /**
   * Uncrypt the specified byte array.
   * 
   * @param pCryptByteArr
   * @return
   */
  public byte[] decrypt(
      final byte[] pCryptByteArr)
  {
    final byte[] decryptByteArr = new byte[ pCryptByteArr.length ];
    for (int i = 0; i < pCryptByteArr.length; i++ )
    {
      byte b = pCryptByteArr[i];
      decryptByteArr[i] = (byte) decryptOne( Str8.byteToInt( b ) );
    }
    return decryptByteArr;
  }
  
  /**
   * Crypt the specified byte buffer
   * with side effects.
   */
  public void cryptIn(
      final Str8Buff pBuff)
  {
    final byte[] cryptByteArr = pBuff.innerByteArr();
    final int length = pBuff.length();
    for (int i = 0; i < length; i++ )
    {
      cryptByteArr[i] = (byte) cryptOne( Str8.byteToInt( cryptByteArr[i] ) );
    }
  }
  
  /**
   * Uncrypt the specified byte buffer
   * with side effects.
   */
  public void decryptIn(
      final Str8Buff pBuff)
  {
    final byte[] cryptByteArr = pBuff.innerByteArr();
    final int length = pBuff.length();
    for (int i = 0; i < length; i++ )
    {
      cryptByteArr[i] = (byte) decryptOne( Str8.byteToInt( cryptByteArr[i] ) );
    }
  }
  
  /**
   * Test.
   * 
   * @param args
   */
  public static void main(
      String[] args )
  {
    final byte[] keyByteArr = "r1U2m3pelsTielzchen".getBytes();
    final Cipher cryptCipher = new Cipher( keyByteArr );
    final Cipher decryptCipher = new Cipher( keyByteArr );
    
    for (int i = 0; i < 255000; i++ )
    {
      //final int baseByte = i;
      final int baseByte = (int) Math.round( Math.random() * 255 );
      System.out.println( "baseByte: " + baseByte );
      
      final int cryptByte = cryptCipher.cryptOne( baseByte );
      System.out.println( "cryptByte: " + cryptByte );

      final int decrByte = decryptCipher.decryptOne( cryptByte );
      System.out.println( "decrByte: " + decrByte );
      
      if ( baseByte != decrByte )
      {
        throw new RuntimeException();
      }
    }

    final byte[] baseByteArr = new byte[ 255000 ];
    for (int i = 0; i < 255000; i++ )
    {
      final byte baseByte = (byte) Math.round( Math.random() * 255 );
      System.out.println( "baseByte: " + baseByte );
      baseByteArr[i] = baseByte;
    }
    final byte[] cryptByteArr = cryptCipher.crypt( baseByteArr );
    final byte[] decryptByteArr = decryptCipher.decrypt( cryptByteArr );
    if ( ! Arrays.equals( decryptByteArr , baseByteArr ) )
    {
      throw new RuntimeException();
    }
    System.out.println( "ok" );
  }

}
