package util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * @author Heiner Kcker
 */
public class SerialUtil
{

  /**
   * Constructor
   */
  private SerialUtil()
  {
    super();
  }

  /** null Byte Code */
  public final static byte NULL_BYTE_CODE = 110;

  /** true Byte Code */
  public final static byte TRUE_BYTE_CODE = 116;

  /** false Byte Code */
  public final static byte FALSE_BYTE_CODE = 102;

  /** byte Byte Code */
  public final static byte BYTE_BYTE_CODE = 98;

  /** short Byte Code */
  public final static byte SHORT_BYTE_CODE = 115;

  /** int Byte Code */
  public final static byte INT_BYTE_CODE = 105;

  /** String Byte Code */
  public final static byte STRING_BYTE_CODE = 83;

  /** short String (length <= Short.MAX_VALUE) Byte Code */
  public final static byte STRING_SHORT_BYTE_CODE = 7;

  /** byte String (length <= Byte.MAX_VALUE) Byte Code */
  public final static byte STRING_BYTE_BYTE_CODE = 8;

  /** Binary Byte Code */
  public final static byte BIN_BYTE_CODE = 66;

  /** long Byte Code */
  private static final byte LONG_BYTE_CODE = 108;

  /** float Byte Code */
  public final static byte FLOAT_BYTE_CODE = 70;

  /** double Byte Code */
  public final static byte DOUBLE_BYTE_CODE = 100;

  /** Array untyped Byte Code */
  public final static byte ARRAY_UNTYPED_BYTE_CODE = 97;

  /** Array typed Byte Code */
  public final static byte ARRAY_TYPED_BYTE_CODE = 65;

  /** Space Byte Code */
  //public final static byte SPACE_BYTE_CODE = 32;
  /** Map Byte Code */
  public final static byte MAP_BYTE_CODE = 77;

  /** Table Byte Code */
  public final static byte TABLE_BYTE_CODE = 84;

  /** GZip Byte Code */
  public final static byte GZIP_BYTE_CODE = 71;

  /** Date Byte Code */
  public final static byte DATE_BYTE_CODE = 68;

  /** Exception Byte Code */
  public final static byte EXCEPTION_BYTE_CODE = 69;
  
  /** Short as Byte (Range: Byte.MIN_VALUE ... Byte.MAX_VALUE) Byte Code */
  public final static byte SHORT_AS_BYTE_BYTE_CODE = 1;
  
  /** Integer as Byte (Range: Byte.MIN_VALUE ... Byte.MAX_VALUE) Byte Code */
  public final static byte INT_AS_BYTE_BYTE_CODE = 2;
  
  /** Integer as Short (Range: Short.MIN_VALUE ... Short.MAX_VALUE) Byte Code */
  public final static byte INT_AS_SHORT_BYTE_BYTE_CODE = 3;
  
  /** Long as Byte (Range: Byte.MIN_VALUE ... Byte.MAX_VALUE) Byte Code */
  public final static byte LONG_AS_BYTE_BYTE_CODE = 4;
  
  /** Long as Short (Range: Short.MIN_VALUE ... Short.MAX_VALUE) Byte Code */
  public final static byte LONG_AS_SHORT_BYTE_BYTE_CODE = 5;
  
  /** Long as Integer (Range: Integer.MIN_VALUE ... Integer.MAX_VALUE) Byte Code */
  public final static byte LONG_AS_INT_BYTE_BYTE_CODE = 6;

  /**
   * serialize an unicode string in UTF-8.
   * 
   * @param pStr
   * @return
   */
  private static byte[] serialize(
      final String pStr )
  {
    final byte[] utf8ByteArr = Str8.strToUtf8( pStr );
    final int dataLength = utf8ByteArr.length;
    final byte[] retByteArr;
    if (dataLength <= Byte.MAX_VALUE)
    {
      retByteArr = new byte[dataLength + 2];
      System.arraycopy( utf8ByteArr , 0 , retByteArr , 2 , dataLength );
      retByteArr[0] = STRING_BYTE_BYTE_CODE;
      retByteArr[1] = (byte) dataLength;
    }
    else if (dataLength <= Short.MAX_VALUE)
    {
      retByteArr = new byte[dataLength + 3];
      System.arraycopy( utf8ByteArr , 0 , retByteArr , 3 , dataLength );
      retByteArr[0] = STRING_SHORT_BYTE_CODE;
      final byte[] lenByteArr = BinUtil.shortToByteArr( (short) dataLength );
      System.arraycopy( lenByteArr , 0 , retByteArr , 1 , 2 );
    }
    else
    {
      retByteArr = new byte[dataLength + 5];
      System.arraycopy( utf8ByteArr , 0 , retByteArr , 5 , dataLength );
      retByteArr[0] = STRING_BYTE_CODE;
      final byte[] lenByteArr = BinUtil.intToByteArr( dataLength );
      System.arraycopy( lenByteArr , 0 , retByteArr , 1 , 4 );
    }

    return retByteArr;
  }

  /**
   * 
   * @param pInByteBuffer
   * @return
   */
  private static String deserializeString(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readInt( pInByteBuffer );
    final byte[] utf8ByteArr = pInByteBuffer.read( length );
    return Str8.utf8ToStr( utf8ByteArr );
  }

  /**
   * 
   * @param pInByteBuffer
   * @return
   */
  private static String deserializeStringShort(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readShort( pInByteBuffer );
    final byte[] utf8ByteArr = pInByteBuffer.read( length );
    return Str8.utf8ToStr( utf8ByteArr );
  }

  /**
   * 
   * @param pInByteBuffer
   * @return
   */
  private static String deserializeStringByte(
      final InByteBuffer pInByteBuffer )
  {
    final int length = pInByteBuffer.read();
    final byte[] utf8ByteArr = pInByteBuffer.read( length );
    return Str8.utf8ToStr( utf8ByteArr );
  }

  /**
   * serialize an binary string.
   * 
   * @param pStr8Buff
   * @return
   */
  private static byte[] serialize(
      final Str8Buff pStr8Buff )
  {
    final int dataLength = pStr8Buff.length();
    final byte[] retByteArr = new byte[dataLength + 5];
    System.arraycopy( pStr8Buff.innerByteArr() , 0 , retByteArr , 5 , dataLength );

    retByteArr[0] = BIN_BYTE_CODE;

    final byte[] lenByteArr = BinUtil.intToByteArr( dataLength );

    System.arraycopy( lenByteArr , 0 , retByteArr , 1 , 4 );

    return retByteArr;
  }

  /**
   * serialize an byte array.
   * 
   * @param pStr8Buff
   * @return
   */
  private static byte[] serialize(
      final byte[] pByteArr )
  {
    final int dataLength = pByteArr.length;
    final byte[] retByteArr = new byte[dataLength + 5];
    System.arraycopy( pByteArr , 0 , retByteArr , 5 , dataLength );

    retByteArr[0] = BIN_BYTE_CODE;

    final byte[] lenByteArr = BinUtil.intToByteArr( dataLength );

    System.arraycopy( lenByteArr , 0 , retByteArr , 1 , 4 );

    return retByteArr;
  }

  /**
   * 
   * @param pInByteBuffer
   * @return
   */
  private static byte[] deserializeBinary(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readInt( pInByteBuffer );
    return pInByteBuffer.read( length );
  }

  /**
   * serialize an byte value (8 Bit).
   * 
   * @param pByte
   * @return
   */
  private static byte[] serialize(
      final byte pByte )
  {
    final byte[] retByteArr = new byte[2];
    retByteArr[0] = BYTE_BYTE_CODE;
    retByteArr[1] = pByte;
    return retByteArr;
  }

  /**
   * deserialize an byte value (8 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static byte deserializeByte(
      final byte[] pByteArr )
  {
    return pByteArr[ 1 ];
  }

  /**
   * deserialize an byte value (8 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static byte deserializeByte(
      final InByteBuffer pInByteBuffer )
  {
    return pInByteBuffer.read();
  }

  /**
   * serialize an short value (16 Bit).
   * 
   * @param pShort
   * @return
   */
  private static byte[] serialize(
      final short pShort )
  {
    final byte[] retByteArr = new byte[3];
    retByteArr[0] = SHORT_BYTE_CODE;
    final byte[] intByteArr = BinUtil.shortToByteArr( pShort );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 2 );
    return retByteArr;
  }

  /**
   * @param pIntValue
   * @return
   */
  private static byte[] serializeIntAsShort(
      int pInt )
  {
    final byte[] retByteArr = new byte[3];
    retByteArr[0] = INT_AS_SHORT_BYTE_BYTE_CODE;
    final byte[] intByteArr = BinUtil.shortToByteArr( (short) pInt );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 2 );
    return retByteArr;
  }

  /**
   * @param pLongValue
   * @return
   */
  private static byte[] serializeLongAsShort(
      long pLong )
  {
    final byte[] retByteArr = new byte[3];
    retByteArr[0] = LONG_AS_SHORT_BYTE_BYTE_CODE;
    final byte[] intByteArr = BinUtil.shortToByteArr( (short) pLong );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 2 );
    return retByteArr;
  }


  /**
   * @param pShortValue
   */
  private static byte[] serializeShortAsByte(
      short pShort )
  {
    final byte[] retByteArr = new byte[2];
    retByteArr[0] = SHORT_AS_BYTE_BYTE_CODE;
    retByteArr[1] = (byte) pShort;
    return retByteArr;
  }

  /**
   * @param pIntValue
   * @return
   */
  private static byte[] serializeIntAsByte(
      int pInt )
  {
    final byte[] retByteArr = new byte[2];
    retByteArr[0] = INT_AS_BYTE_BYTE_CODE;
    retByteArr[1] = (byte) pInt;
    return retByteArr;
  }


  /**
   * @param pLongValue
   * @return
   */
  private static byte[] serializeLongAsByte(
      long pLong )
  {
    final byte[] retByteArr = new byte[2];
    retByteArr[0] = LONG_AS_BYTE_BYTE_CODE;
    retByteArr[1] = (byte) pLong;
    return retByteArr;
  }



  /**
   * deserialize an byte value (8 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static short deserializeShort(
      final byte[] pByteArr )
  {
    final byte[] intByteArr = new byte[2];
    System.arraycopy( pByteArr , 1 , intByteArr , 0 , 2 );
    return BinUtil.byteArrToShort( intByteArr );
  }

  /**
   * deserialize an byte value (8 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static short deserializeShort(
      final InByteBuffer pInByteBuffer )
  {
    return BinUtil.readShort( pInByteBuffer );
  }

  /**
   * serialize an int value (32 Bit).
   * 
   * @param pInt
   * @return
   */
  private static byte[] serialize(
      final int pInt )
  {
    final byte[] retByteArr = new byte[5];
    retByteArr[0] = INT_BYTE_CODE;
    final byte[] intByteArr = BinUtil.intToByteArr( pInt );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 4 );
    return retByteArr;
  }

  /**
   * @param pLongValue
   * @return
   */
  private static byte[] serializeLongAsInt(
      long pLong )
  {
    final byte[] retByteArr = new byte[5];
    retByteArr[0] = LONG_AS_INT_BYTE_BYTE_CODE;
    final byte[] intByteArr = BinUtil.intToByteArr( (int) pLong );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 4 );
    return retByteArr;
  }


  /**
   * deserialize an int value (32 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static int deserializeInt(
      final byte[] pByteArr )
  {
    final byte[] intByteArr = new byte[4];
    System.arraycopy( pByteArr , 1 , intByteArr , 0 , 4 );
    return BinUtil.byteArrToInt( intByteArr );
  }

  /**
   * deserialize an int value (32 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static int deserializeInt(
      final InByteBuffer pInByteBuffer )
  {
    return BinUtil.readInt( pInByteBuffer );
  }

  /**
   * serialize an long value (64 Bit).
   * 
   * @param pInt
   * @return
   */
  private static byte[] serialize(
      final long pInt )
  {
    final byte[] retByteArr = new byte[9];
    retByteArr[0] = LONG_BYTE_CODE;
    final byte[] intByteArr = BinUtil.longToByteArr( pInt );
    System.arraycopy( intByteArr , 0 , retByteArr , 1 , 8 );
    return retByteArr;
  }

  /**
   * deserialize an long value (64 Bit).
   * 
   * @param pByteArr
   * @return
   */
  private static long deserializeLong(
      final byte[] pByteArr )
  {
    final byte[] longByteArr = new byte[8];
    System.arraycopy( pByteArr , 1 , longByteArr , 0 , 8 );
    return BinUtil.byteArrToLong( longByteArr );
  }

  /**
   * deserialize an long value (64 Bit).
   * 
   * @param pInByteBuffer
   * @return
   */
  private static long deserializeLong(
      final InByteBuffer pInByteBuffer )
  {
    return BinUtil.readLong( pInByteBuffer );
  }

  /**
   * 
   * @param pFloat
   * @return
   */
  private static byte[] serialize(
      final float pFloat )
  {
    final byte[] retByteArr = new byte[5];

    retByteArr[0] = FLOAT_BYTE_CODE;

    final byte[] longByteArr = BinUtil.floatToByteArr( pFloat );

    System.arraycopy( longByteArr , 0 , retByteArr , 1 , 4 );

    return retByteArr;
  }

  /**
   * 
   * @param pByteArr
   * @return
   */
  private static float deserializeFloat(
      final byte[] pByteArr )
  {
    final byte[] intByteArr = new byte[4];

    System.arraycopy( pByteArr , 1 , intByteArr , 0 , 4 );

    final float retFloat = BinUtil.byteArrToFloat( intByteArr );
    return retFloat;
  }

  /**
   * 
   * @param pByteArr
   * @return
   */
  private static float deserializeFloat(
      final InByteBuffer pInByteBuffer )
  {
    return BinUtil.readFloat( pInByteBuffer );
  }

  /**
   * 
   * @param pDouble
   * @return
   */
  private static byte[] serialize(
      final double pDouble )
  {
    final byte[] retByteArr = new byte[9];

    retByteArr[0] = DOUBLE_BYTE_CODE;

    final byte[] longByteArr = BinUtil.doubleToByteArr( pDouble );

    System.arraycopy( longByteArr , 0 , retByteArr , 1 , 8 );

    return retByteArr;
  }

  /**
   * 
   * @param pByteArr
   * @return
   */
  private static double deserializeDouble(
      final byte[] pByteArr )
  {
    final byte[] longByteArr = new byte[8];

    System.arraycopy( pByteArr , 1 , longByteArr , 0 , 8 );

    final double retDouble = BinUtil.byteArrToDouble( longByteArr );
    return retDouble;
  }

  /**
   * 
   * @param pByteArr
   * @return
   */
  private static double deserializeDouble(
      final InByteBuffer pInByteBuffer )
  {
    return BinUtil.readDouble( pInByteBuffer );
  }

  /**
   * 
   * @param pDate
   * @return
   */
  private static byte[] serialize(
      final Date pDate )
  {
    final byte[] retByteArr = new byte[9];

    retByteArr[0] = DATE_BYTE_CODE;

    final long dateBinaryLong = pDate.getTime();

    final byte[] longByteArr = BinUtil.longToByteArr( dateBinaryLong );

    System.arraycopy( longByteArr , 0 , retByteArr , 1 , 8 );

    return retByteArr;
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  private static Date deserializeDate(
      final InByteBuffer pInByteBuffer )
  {
    final long time = BinUtil.readLong( pInByteBuffer );
    return new Date( time );
  }

  /**
   * 
   * @param pByteArr
   * @return
   */
  private static Date deserializeDate(
      final byte[] pByteArr )
  {
    final byte[] longByteArr = new byte[8];

    System.arraycopy( pByteArr , 1 , longByteArr , 0 , 8 );

    final long dateBinaryLong = BinUtil.byteArrToLong( longByteArr );

    return new Date( dateBinaryLong );
  }

  /**
   * 
   * @param pObjectArr
   * @return
   */
  private static byte[] serialize(
      final Object[] pObjectArr )
  {
    return serialize( Arrays.asList( pObjectArr ) );
  }

  /**
   * 
   * @param pList
   * @return
   */
  private static byte[] serialize(
      final List pList )
  {
    //    if ( pList.size() > Short.MAX_VALUE)
    //    {
    //      throw new IllegalArgumentException( "list is to big: " + pList.size() );
    //    }

    final Str8Buff buff = new Str8Buff( 2 * 1024 );

    // header
    buff.append( ARRAY_UNTYPED_BYTE_CODE );
    //buff.append( SPACE_BYTE_CODE );
    //buff.append( BinUtil.shortToByteArr( (short) pList.size() ) );
    buff.append( BinUtil.intToByteArr( pList.size() ) );

    for (int i = 0; i < pList.size(); i++ )
    {
      final Object element = pList.get( i );

      appendElement( buff , element , i );
    }
    byte[] retByteArr = buff.bytes();

    return retByteArr;
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  //  private static Object[] deserializeArrayUntyped(
  //      final InByteBuffer pInByteBuffer )
  //  {
  //    final int length = BinUtil.readInt( pInByteBuffer );
  //    final Object[] retObjArr = new Object[ length ];
  //    for (int i = 0; i < length; i++ )
  //    {
  //      retObjArr[ i ] = deserialize( pInByteBuffer );
  //    }
  //    return retObjArr;
  //  }
  /**
   * hier mit ArrayList, weil equals auf Array nicht funktioniert.
   * 
   * 
   * @param pInByteBuffer
   * @return
   */
  private static ArrayList deserializeArrayUntyped(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readInt( pInByteBuffer );
    final ArrayList<Object> retArrList = new ArrayList<Object>( length );
    for (int i = 0; i < length; i++ )
    {
      retArrList.add( deserialize( pInByteBuffer ) );
    }
    return retArrList;
  }

  /**
   * 
   * @param pMap
   * @return
   */
  private static byte[] serialize(
      final Map pMap )
  {
    //    if ( pMap.size() > Short.MAX_VALUE)
    //    {
    //      throw new IllegalArgumentException( "map is to big: " + pMap.size() );
    //    }

    final Str8Buff buff = new Str8Buff( 2 * 1024 );

    // header
    buff.append( MAP_BYTE_CODE );
    buff.append( BinUtil.intToByteArr( pMap.size() ) );
    //buff.append( BinUtil.shortToByteArr( (short) pMap.size() ) );

    for (Iterator iter = pMap.entrySet().iterator(); iter.hasNext();)
    {
      Map.Entry entry = (Map.Entry) iter.next();
      appendElement( buff , entry.getKey() , 0 );
      appendElement( buff , entry.getValue() , 0 );
    }

    byte[] retByteArr = buff.bytes();

    return retByteArr;
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  private static HashMap deserializeMap(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readInt( pInByteBuffer );
    final HashMap<Object, Object> retMap = new HashMap<Object, Object>( length );
    for (int i = 0; i < length; i++ )
    {
      final Object key = deserialize( pInByteBuffer );
      final Object value = deserialize( pInByteBuffer );
      retMap.put( key , value );
    }
    return retMap;
  }

  /**
   * 
   * @param pTable
   * @return
   */
  private static byte[] serialize(
      final Table pTable )
  {
    final int size = pTable.size();
    final Str8Buff buff = new Str8Buff( size * 256 );

    buff.append( TABLE_BYTE_CODE );

    buff.append( serialize( pTable.getColumnNameArr() ) );

    buff.append( BinUtil.intToByteArr( size ) );
    
    for (int i = 0, n = size; i < n; i++ )
    {
      Object[] rowObjArr = pTable.getRow( i );
      for (int j = 0; j < rowObjArr.length; j++ )
      {
        final Object value = rowObjArr[j];
        appendElement( buff , value , 0 );
      }
    }

    return buff.bytes();
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  private static Table deserializeTable(
      final InByteBuffer pInByteBuffer )
  {
    //final Object[] colNameObjArr = (Object[]) deserialize( pInByteBuffer );
    final Object[] colNameObjArr = toObjArr( (ArrayList) deserialize( pInByteBuffer ) );
    final int columnCount = colNameObjArr.length;
    final String[] colNameArr = new String[columnCount];
    System.arraycopy( colNameObjArr , 0 , colNameArr , 0 , columnCount );

    final int length = BinUtil.readInt( pInByteBuffer );
    final Table retTable = new Table( colNameArr , length );
    for (int i = 0; i < length; i++ )
    {
      for (int j = 0; j < columnCount; j++ )
      {
        final Object value = deserialize( pInByteBuffer );
        retTable.set( i , j , value );
      }
    }
    return retTable;
  }

  /**
   * @param pExceptionHolder
   * @return
   */
  private static byte[] serialize(
      final ExceptionHolder pExceptionHolder )
  {
    final Str8Buff subBuff = new Str8Buff( 1024 );
    final byte[] nestedCounterByteArr = new byte[1];
    subSerialize( pExceptionHolder , nestedCounterByteArr, subBuff );
    final Str8Buff retBuff = new Str8Buff( subBuff.length() + 2 );
    retBuff.append( EXCEPTION_BYTE_CODE );
    retBuff.append( nestedCounterByteArr );
    retBuff.append( subBuff.bytes() );
    return retBuff.bytes();
  }

  /**
   * recursiv
   * @param pExceptionHolder
   * @param pNestedCounterByteArr
   * @param pSubBuff
   */
  private static void subSerialize(
      final ExceptionHolder pExceptionHolder ,
      final byte[] pNestedCounterByteArr ,
      final Str8Buff pSubBuff )
  {
    pNestedCounterByteArr[0]++;
    if ( pNestedCounterByteArr[0] < 0 )
    {
      // over 127 nested exceptions, number rolls to < 0
      return;
    }
    appendElement( pSubBuff , pExceptionHolder.className , 0 );
    appendElement( pSubBuff , pExceptionHolder.message , 0 );
    appendElement( pSubBuff , pExceptionHolder.stacktrace , 0 );
    final ExceptionHolder causeExceptionHolder = pExceptionHolder.causeExceptionHolder;
    if ( causeExceptionHolder != null )
    {
      subSerialize( causeExceptionHolder , pNestedCounterByteArr , pSubBuff );
    }
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  private static ExceptionHolder deserializeException(
      InByteBuffer pInByteBuffer )
  {
    final byte nestedCounter = pInByteBuffer.read();
    return subDeserializeException( pInByteBuffer, nestedCounter );
  }

  /**
   * recursiv
   * @param pInByteBuffer
   * @param pNestedCounter
   * @return
   */
  private static ExceptionHolder subDeserializeException(
      final InByteBuffer pInByteBuffer ,
      byte pNestedCounter )
  {
    if ( pNestedCounter <= 0 )
    {
      return null;
    }
    return new ExceptionHolder( 
        (String) deserialize( pInByteBuffer) , 
        (String) deserialize( pInByteBuffer) , 
        (String) deserialize( pInByteBuffer) , 
        subDeserializeException( pInByteBuffer , --pNestedCounter ) );
  }

  /**
   * Serialize the specified
   * Object by according the type 
   * and append it to the specified
   * Buffer.
   * 
   * @param pBuff
   * @param pElement
   * @param pIndex
   */
  public static void appendElement(
      final Str8Buff pBuff ,
      final Object pElement ,
      final int pIndex )
  {
    final byte[] retElemByteArr = serialize( pElement , pIndex );
    
    pBuff.append( retElemByteArr );
  }

  /**
   * Serialize the specified
   * Object by according the type.
   * 
   * @param pElement
   * @param pIndex only for error message
   * @return
   */
  public static byte[] serialize(
      final Object pElement ,
      final int pIndex )
  {
    final byte[] retElemByteArr;
    if (pElement == null)
    {
      retElemByteArr = new byte[]{ NULL_BYTE_CODE };
    }
    else if (pElement instanceof Boolean)
    {
      if (( (Boolean) pElement ).booleanValue())
      {
        retElemByteArr = new byte[]{ TRUE_BYTE_CODE };
      }
      else
      {
        retElemByteArr = new byte[]{ FALSE_BYTE_CODE };
      }
    }
    else if (pElement instanceof byte[])
    {
      retElemByteArr = serialize( (byte[]) pElement ) ;
    }
    else if (pElement instanceof Str8Buff)
    {
      retElemByteArr = serialize( (Str8Buff) pElement ) ;
    }
    else if (pElement instanceof String)
    {
      retElemByteArr = serialize( (String) pElement ) ;
    }
    else if (pElement instanceof Integer )
    {
      final int intValue = ( (Integer) pElement ).intValue();
      if ( intValue >= Byte.MIN_VALUE && intValue <= Byte.MAX_VALUE )
      {
        retElemByteArr = serializeIntAsByte( intValue );
      }
      else if ( intValue >= Short.MIN_VALUE && intValue <= Short.MAX_VALUE )
      {
        retElemByteArr = serializeIntAsShort( intValue );
      }
      else
      {
        retElemByteArr = serialize( intValue ) ;
      }
    }
    else if (pElement instanceof Long)
    {
      final long longValue = ( (Long) pElement ).longValue();
      if ( longValue >= Byte.MIN_VALUE && longValue <= Byte.MAX_VALUE )
      {
        retElemByteArr = serializeLongAsByte( longValue );
      }
      else if ( longValue >= Short.MIN_VALUE && longValue <= Short.MAX_VALUE )
      {
        retElemByteArr = serializeLongAsShort( longValue );
      }
      else if ( longValue >= Integer.MIN_VALUE && longValue <= Integer.MAX_VALUE )
      {
        retElemByteArr = serializeLongAsInt( longValue );
      }
      else
      {
        retElemByteArr = serialize( longValue ) ;
      }
    }
    else if (pElement instanceof Byte)
    {
      retElemByteArr = serialize( ( (Number) pElement ).byteValue() ) ;
    }
    else if ( pElement instanceof Short)
    {
      final short shortValue = ( (Short) pElement ).shortValue();
      if ( shortValue >= Byte.MIN_VALUE && shortValue <= Byte.MAX_VALUE )
      {
        retElemByteArr = serializeShortAsByte( shortValue );
      }
      else
      {
        retElemByteArr = serialize( shortValue ) ;
      }
    }
    else if (pElement instanceof Float)
    {
      retElemByteArr = serialize( ( (Number) pElement ).floatValue() ) ;
    }
    else if (pElement instanceof Number)
    {
      retElemByteArr = serialize( ( (Number) pElement ).doubleValue() ) ;
    }
    else if (pElement instanceof List)
    {
      retElemByteArr = serialize( (List) pElement ) ;
    }
    else if (pElement instanceof Map)
    {
      retElemByteArr = serialize( (Map) pElement ) ;
    }
    else if (pElement instanceof Object[])
    {
      retElemByteArr = serialize( (Object[]) pElement ) ;
    }
    else if (pElement instanceof Date)
    {
      retElemByteArr = serialize( (Date) pElement ) ;
    }
    else if (pElement instanceof Table)
    {
      retElemByteArr = serialize( (Table) pElement ) ;
    }
    else if (pElement instanceof GzipHolder)
    {
      retElemByteArr = serialize( ( GzipHolder ) pElement ) ;
    }
    else if (pElement instanceof ExceptionHolder)
    {
      retElemByteArr = serialize( ( ExceptionHolder ) pElement ) ;
    }
    else if (pElement instanceof Exception)
    {
      retElemByteArr = serialize( new ExceptionHolder( (Throwable) pElement ) );
    }
    else
    {
      throw new IllegalArgumentException( "["
          + pIndex + "]: " + pElement.getClass().getName() + ":" + pElement );
    }
    return retElemByteArr;
  }

  /**
   * Wrap the specified data (byte array)
   * in an GZip envelope.
   * 
   * @param pByteArr
   * @return
   */
  private static byte[] wrapGzip(
      final byte[] pByteArr )
  {
    final byte[] gzipByteArr = GzipUtil.gzip( pByteArr );
    final byte[] retByteArr = new byte[gzipByteArr.length + 5];
    System.arraycopy( gzipByteArr , 0 , retByteArr , 5 , gzipByteArr.length );
    retByteArr[0] = GZIP_BYTE_CODE;
    final byte[] lengthByteArr = BinUtil.intToByteArr( gzipByteArr.length );
    System.arraycopy( lengthByteArr , 0 , retByteArr , 1 , 4 );
    return retByteArr;
  }

  /**
   * @param pInByteBuffer
   * @return
   */
  private static InByteBuffer unwrapGzip(
      final InByteBuffer pInByteBuffer )
  {
    final int length = BinUtil.readInt( pInByteBuffer );
    final byte[] gzipByteArr = pInByteBuffer.read( length );
    final byte[] gunzipByteArr = GzipUtil.gunzip( gzipByteArr );
    return new InByteBuffer( gunzipByteArr );
  }

  /**
   * 
   * @param pDate
   * @return
   */
  private static byte[] serialize(
      final GzipHolder pGzipHolder )
  {
    // in temporren (sub) Puffer serialisieren
    //und dann in einen GZIP-Umschlag verpacken
    final Str8Buff subBuff = new Str8Buff( 1024 );
    appendElement( subBuff, pGzipHolder.objToCompress , 0 );
    return wrapGzip( subBuff.bytes() );
  }

  /**
   * 
   * @param pInByteBuffer
   * @return
   */
  public static Object deserialize(
      final InByteBuffer pInByteBuffer )
  {
    byte codeByte = pInByteBuffer.read();
    switch (codeByte)
    {
      case GZIP_BYTE_CODE:
        final InByteBuffer gunzipInByteBuffer = unwrapGzip( pInByteBuffer );
        return deserialize( gunzipInByteBuffer );

      case NULL_BYTE_CODE:
        return null;

      case TRUE_BYTE_CODE:
        return Boolean.TRUE;

      case FALSE_BYTE_CODE:
        return Boolean.FALSE;

      case BYTE_BYTE_CODE:
        return deserializeByte( pInByteBuffer );

      case SHORT_BYTE_CODE:
        return deserializeShort( pInByteBuffer );

      case INT_BYTE_CODE:
        return deserializeInt( pInByteBuffer );

      case LONG_BYTE_CODE:
        return deserializeLong( pInByteBuffer );

      case FLOAT_BYTE_CODE:
        return deserializeFloat( pInByteBuffer );

      case DOUBLE_BYTE_CODE:
        return deserializeDouble( pInByteBuffer );

      case STRING_BYTE_CODE:
        return deserializeString( pInByteBuffer );

      case STRING_SHORT_BYTE_CODE:
        return deserializeStringShort( pInByteBuffer );

      case STRING_BYTE_BYTE_CODE:
        return deserializeStringByte( pInByteBuffer );

      case BIN_BYTE_CODE:
        return deserializeBinary( pInByteBuffer );

      case DATE_BYTE_CODE:
        return deserializeDate( pInByteBuffer );

      case ARRAY_UNTYPED_BYTE_CODE:
        return deserializeArrayUntyped( pInByteBuffer );

      case MAP_BYTE_CODE:
        return deserializeMap( pInByteBuffer );

      case TABLE_BYTE_CODE:
        return deserializeTable( pInByteBuffer );

      case EXCEPTION_BYTE_CODE:
        return deserializeException( pInByteBuffer );

      case SHORT_AS_BYTE_BYTE_CODE:
        //return deserializeShortAsByte( pInByteBuffer );
        return (short) deserializeByte( pInByteBuffer );

      case INT_AS_BYTE_BYTE_CODE:
        //return deserializeIntAsByte( pInByteBuffer );
        return (int) deserializeByte( pInByteBuffer );

      case INT_AS_SHORT_BYTE_BYTE_CODE:
        //return deserializeIntAsShort( pInByteBuffer );
        return (int) deserializeShort( pInByteBuffer );

      case LONG_AS_BYTE_BYTE_CODE:
        //return deserializeLongAsByte( pInByteBuffer );
        return (long) deserializeByte( pInByteBuffer );

      case LONG_AS_SHORT_BYTE_BYTE_CODE:
        //return deserializeLongAsShort( pInByteBuffer );
        return (long) deserializeShort( pInByteBuffer );

      case LONG_AS_INT_BYTE_BYTE_CODE:
        //return deserializeLongAsInt( pInByteBuffer );
        return (long) deserializeInt( pInByteBuffer );

      default:
        throw new IllegalArgumentException( "unsupported data type code "
            + codeByte + " " + (char) Str8.byteToInt( codeByte ) );
    //break;
    }
  }

  /**
   * Serialize the specified count
   * of values.
   * 
   * @param pRespObjArr
   * @return
   */
  public static Str8Buff serializeMulti(
      final Object[] pRespObjArr )
  {
    final Str8Buff respValuesBuff = new Str8Buff( 1024 );
    for (int i = 0; i < pRespObjArr.length; i++ )
    {
      final Object respValue = pRespObjArr[i];
      appendElement( respValuesBuff, respValue, i );
    }
    return respValuesBuff;
  }

  /**
   * Deserialize the specified count
   * of values.
   * 
   * @param pInByteBuffer
   * @param pValueCount
   * @return
   */
  public static Object[] deserializeMulti(
      final InByteBuffer pInByteBuffer ,
      int pValueCount )
  {
System.out.println( "pValueCount: " + pValueCount );
    final Object[] valueObjArr = new Object[ pValueCount ];
    for (int i = 0; i < pValueCount; i++ )
    {
      valueObjArr[ i ] = deserialize( pInByteBuffer );
    }
    
    // check whether all values computed, only for develop
    if ( ! pInByteBuffer.empty() )
    {
      throw new RuntimeException();
    }
    
    return valueObjArr;
  }

  /**
   * 
   * @param pObjArr
   * @return
   */
  public static final ArrayList<Object> toArrList(
      final Object[] pObjArr )
  {
    if (pObjArr == null)
    {
      return null;
    }
    final ArrayList<Object> arrList = new ArrayList<Object>( pObjArr.length );
    for (int i = 0; i < pObjArr.length; i++ )
    {
      arrList.add( pObjArr[i] );
    }
    return arrList;
  }

  /**
   * Create a Arraylist from varargs.
   * 
   * @param pObjArr
   * @return
   */
  public static ArrayList<Object> newArrList(
      final Object... pObjArr )
  {
    final ArrayList<Object> list = new ArrayList<Object>( pObjArr.length );
    for (Object value : pObjArr)
    {
      list.add( value );
    }
    return list;
  }
  
  /**
   * @param pBs
   * @return
   */
  public static ArrayList toArrList(
      byte[] pByteArr )
  {
    if ( pByteArr == null )
    {
      return null;
    }
    final ArrayList<Object> list = new ArrayList<Object>( pByteArr.length );
    for (int i = 0; i < pByteArr.length; i++ )
    {
      list.add( pByteArr[i] );
    }
    return list;
  }

  /**
   * @param pObject
   * @return
   */
  public static final Object[] toObjArr(
      List pList )
  {
    if ( pList == null )
    {
      return null;
    }
    return (Object[]) pList.toArray( new Object[ pList.size() ] );
  }

  /**
   * Test.
   * 
   * @param args
   */
  public static void main(
      final String[] args )
  {
    if (false)
    {
      // Test serialize(int) und deserializeInt
      for (int i = Short.MIN_VALUE; i < Short.MAX_VALUE; i++ )
      {
        //System.out.println( i );
        byte[] serialByteArr = serialize( i );
        int deserialShort = deserializeInt( serialByteArr );
        if (deserialShort != i)
        {
          throw new RuntimeException( ""
              + i );
        }
      }
    }

    //---------------------------------------
    ArrayList arr = toArrList( new Object[] {
        "arr1", "arr2", "arr3", "arr4", "arr5", 1, 2L, null } );

    final byte[] arrByteArr = serialize( arr );

    System.out.println( new String( arrByteArr ) );

    final ArrayList deserializedArr = (ArrayList) deserialize( new InByteBuffer( arrByteArr ) );

    System.out.println( CollectionUtil.toStr( deserializedArr ) );

    if ( !arr.equals( deserializedArr ))
    {
      throw new RuntimeException();
    }

    //---------------------------------------
    final Table table = new Table( new String[]{ "Nr" , "Name" , "Size" } );
    for (int i = 0; i < 3; i++ )
    {
      table.add( new Object[]{ i , "Name " + i , i * 100.0  } );
    }

    final byte[] tableByteArr = serialize( table );

    System.out.println( new String( tableByteArr ) );

    final Table deserializedTable = (Table) deserialize( new InByteBuffer( tableByteArr ) );

    System.out.println( deserializedTable );

    if ( !table.equals( deserializedTable ))
    {
      throw new RuntimeException();
    }

    //---------------------------------------
    
    final HashMap<Object, Object> map = new HashMap<Object, Object>();
    map.put( "a" , "a1" );
    map.put( "b" , "b1" );
    map.put( "c" , "c1" );
    map.put( "" , "" );
    map.put( "" , "" );
    map.put( "$" , "" );
    map.put( 1 , 2 );
    map.put( 3 , 4L );
    map.put( (byte) 5 , (short) 6 );
    map.put( 5000000L , 6000000L );
    map.put( "true" , true );
    map.put( "false" , false );
    map.put( 3.0 , 4.0 );
    map.put( 3.3 , 4.4 );
    map.put( 3.4F , 4.5F );
    map.put( "array" , toArrList( new Object[] {
        null, "arr1", "arr2", "arr3", "arr4", "arr5" } ) );

    map.put( new Date() , new Date() );
    
    
    // TODO byte[]
    //map.put( new byte[]{ 1,2 ,3,4,5,6  } , new byte[]{ -1,-2,-3,-4,-5,-6  } );
    
    map.put( "table" , table );

    System.out.println( CollectionUtil.toStr( map ) );

    final byte[] mapByteArr = serialize( map );
    //final byte[] mapByteArr = serialize( new GzipHolder( map ) );

    System.out.println( new String( mapByteArr ) );

    final HashMap deserializedMap = (HashMap) deserialize( new InByteBuffer( mapByteArr ) );

    System.out.println( CollectionUtil.toStr( deserializedMap ) );

    if ( !map.equals( deserializedMap ))
    {
      throw new RuntimeException();
    }

    
    final ExceptionHolder exceptionHolder = new ExceptionHolder( new RuntimeException( "xxx" ) );
    final byte[] excByteArr = serialize( exceptionHolder );
System.out.println( "excByteArr: " + new String( excByteArr ) );
    final ExceptionHolder deserialExceptionHolder = (ExceptionHolder) deserialize( new InByteBuffer( excByteArr ) );
    if ( !exceptionHolder.equals( deserialExceptionHolder ))
    {
      throw new RuntimeException();
    }

    // ----------------------------
    
    final String bigNrStr = "14501164060000";
    
    final double bigNrDouble = Double.valueOf( bigNrStr );
    
    final byte[] bigNrDoubleByteArr = serialize( bigNrDouble );
    
    System.out.println( toArrList( bigNrDoubleByteArr ) );

    if ( bigNrDouble != ( (Double) deserialize( new InByteBuffer( bigNrDoubleByteArr ) ) ).doubleValue() ) 
    {
      throw new RuntimeException();
    }
    
    final long bigNrLong = Long.valueOf( bigNrStr );
    
    final byte[] bigNrLongByteArr = serialize( bigNrLong );
    
    System.out.println( toArrList( bigNrLongByteArr ) );
    
    if ( bigNrLong != ( (Long) deserialize( new InByteBuffer( bigNrLongByteArr ) ) ).longValue() ) 
    {
      throw new RuntimeException();
    }

    // short as byte
    short shortAsByte = Byte.MIN_VALUE;
    testShortAsByte( shortAsByte );
    shortAsByte = Byte.MAX_VALUE;
    testShortAsByte( shortAsByte );
    shortAsByte = 0;
    testShortAsByte( shortAsByte );
    
    // int as byte
    int intAsByte = Byte.MIN_VALUE;
    testIntAsByte( intAsByte );
    intAsByte = Byte.MAX_VALUE;
    testIntAsByte( intAsByte );
    intAsByte = 0;
    testIntAsByte( intAsByte );
    
    // int as short
    int intAsShort = Short.MIN_VALUE;
    testIntAsShort( intAsShort );
    intAsShort = Short.MAX_VALUE;
    testIntAsShort( intAsShort );
    
    // long as byte
    long longAsByte = Byte.MIN_VALUE;
    testLongAsByte( longAsByte );
    longAsByte = Byte.MAX_VALUE;
    testLongAsByte( longAsByte );
    longAsByte = 0;
    testLongAsByte( longAsByte );
    
    // long as short
    long longAsShort = Short.MIN_VALUE;
    testLongAsShort( longAsShort );
    longAsShort = Short.MAX_VALUE;
    testLongAsShort( longAsShort );
    
    // long as int
    long longAsInteger = Integer.MIN_VALUE;
    testLongAsInteger( longAsInteger );
    longAsInteger = Integer.MAX_VALUE;
    testLongAsInteger( longAsInteger );

    // string byte
    char[] char8Arr = new char[ Byte.MAX_VALUE ];
    Arrays.fill( char8Arr , '8' );
    String a8 = new String( char8Arr );
    System.out.println( "a8: " + a8 );
    byte[] a8ByteArr = serialize( a8 , 0 );
    System.out.println( toArrList( a8ByteArr ) );
    if ( ! a8.equals( deserialize( new InByteBuffer( a8ByteArr) ) ) )
    {
      throw new RuntimeException();
    }

    // string short
    //char[] char16Arr = new char[ Short.MAX_VALUE ];
    char[] char16Arr = new char[ Byte.MAX_VALUE * 2 ];
    Arrays.fill( char16Arr , '6' );
    String a16 = new String( char16Arr );
    System.out.println( "a16: " + a16 );
    byte[] a16ByteArr = serialize( a16 , 0 );
    System.out.println( toArrList( a16ByteArr ) );
    if ( ! a16.equals( deserialize( new InByteBuffer( a16ByteArr) ) ) )
    {
      throw new RuntimeException();
    }
    

    
    System.out.println( "ok" );
  }

  /**
   * @param bigNrLongByteArr
   * @param shortAsByte
   */
  private static void testShortAsByte(
      final short shortAsByte )
  {
    System.out.println( "shortAsByte: " + shortAsByte );
    final byte[] shortAsByteByteArr = serialize( new Short( shortAsByte ) , 0 );
    System.out.println( "shortAsByteByteArr: " + toArrList( shortAsByteByteArr ) );
    if ( shortAsByte != ( (Short) deserialize( new InByteBuffer( shortAsByteByteArr ) ) ).shortValue() ) 
    {
      throw new RuntimeException();
    }
  }

  /**
   * @param bigNrLongByteArr
   * @param intAsByte
   */
  private static void testIntAsByte(
      final int intAsByte )
  {
    System.out.println( "intAsByte: " + intAsByte );
    final byte[] intAsByteByteArr = serialize( new Integer( intAsByte ) , 0 );
    System.out.println( "intAsByteByteArr: " + toArrList( intAsByteByteArr ) );
    if ( intAsByte != ( (Integer) deserialize( new InByteBuffer( intAsByteByteArr ) ) ).intValue() ) 
    {
      throw new RuntimeException();
    }
  }

  /**
   * @param bigNrLongByteArr
   * @param intAsByte
   */
  private static void testIntAsShort(
      final int intAsShort )
  {
    System.out.println( "intAsShort: " + intAsShort );
    final byte[] intAsShortByteArr = serialize( new Integer( intAsShort ) , 0 );
    System.out.println( "intAsShortByteArr: " + toArrList( intAsShortByteArr ) );
    if ( intAsShort != ( (Integer) deserialize( new InByteBuffer( intAsShortByteArr ) ) ).intValue() ) 
    {
      throw new RuntimeException();
    }
  }

  /**
   * @param bigNrLongByteArr
   * @param intAsByte
   */
  private static void testLongAsByte(
      final long longAsByte )
  {
    System.out.println( "longAsByte: " + longAsByte );
    final byte[] longAsByteByteArr = serialize( new Long( longAsByte ) , 0 );
    System.out.println( "longAsByteByteArr: " + toArrList( longAsByteByteArr ) );
    if ( longAsByte != ( (Long) deserialize( new InByteBuffer( longAsByteByteArr ) ) ).longValue() ) 
    {
      throw new RuntimeException();
    }
  }

  /**
   * @param bigNrLongByteArr
   * @param longAsByte
   */
  private static void testLongAsShort(
      final long longAsShort )
  {
    System.out.println( "longAsShort: " + longAsShort );
    final byte[] longAsShortByteArr = serialize( new Long( longAsShort ) , 0 );
    System.out.println( "longAsShortByteArr: " + toArrList( longAsShortByteArr ) );
    if ( longAsShort != ( (Long) deserialize( new InByteBuffer( longAsShortByteArr ) ) ).longValue() ) 
    {
      throw new RuntimeException();
    }
  }

  /**
   * @param bigNrLongByteArr
   * @param longAsByte
   */
  private static void testLongAsInteger(
      final long longAsInteger )
  {
    System.out.println( "longAsInteger: " + longAsInteger );
    final byte[] longAsIntegerByteArr = serialize( new Long( longAsInteger ) , 0 );
    System.out.println( "longAsIntegerByteArr: " + toArrList( longAsIntegerByteArr ) );
    if ( longAsInteger != ( (Long) deserialize( new InByteBuffer( longAsIntegerByteArr ) ) ).longValue() ) 
    {
      throw new RuntimeException();
    }
  }

}
