package util;

import java.util.ArrayList;
import java.util.Arrays;

/**
 * A table structure with named but non typed
 * columns.
 * 
 * @author Heiner Kcker
 */
public class Table
{
  /**
   * Column Name Array.
   */
  private final String[] columnNameArr;

  /**
   * List of rows.
   */
  private final ArrayList<Object[]> rowList;

  /**
   * Constructor
   */
  public Table(final String[] pColumnNameArr)
  {
    this( pColumnNameArr , 0 );
  }

  /**
   * Constructor with specified count of rows.
   * 
   * @param pColNameArr
   * @param pLength count of rows
   */
  public Table(final String[] pColumnNameArr, final int pLength)
  {
    if (pColumnNameArr == null
        || pColumnNameArr.length < 1)
    {
      throw new IllegalArgumentException();
    }
    this.columnNameArr = pColumnNameArr;
    if (pLength < 1)
    {
      // null rows
      this.rowList = new ArrayList<Object[]>();
    }
    else
    {
      // add the specified count of rows
      this.rowList = new ArrayList<Object[]>( pLength );
      for (int i = 0; i < pLength; i++ )
      {
        this.rowList.add( new Object[ this.columnNameArr.length ] );
      }
    }
  }

  /**
   * get the count of rows.
   * 
   * @return
   */
  public int rowCount()
  {
    return this.rowList.size();
  }

  /**
   * get the count of rows.
   * 
   * @return
   */
  public int size()
  {
    return this.rowList.size();
  }

  /**
   * get the count of columns.
   * @return
   */
  public int columnCount()
  {
    return this.columnNameArr.length;
  }

  /**
   * set the entire row at the
   * specified position.
   * 
   * @param pRowNumber
   * @param pRowValueArr
   */
  public void set(
      final int pRowNumber ,
      final Object[] pRowValueArr )
  {
    if (pRowValueArr.length != this.columnNameArr.length)
    {
      throw new IllegalArgumentException( "param array length "
          + pRowValueArr.length + " is not equal to colum count: " + this.columnNameArr.length );
    }
    this.rowList.set( pRowNumber , pRowValueArr );
  }

  /**
   * set the value at the specified
   * position.
   * 
   * @param pRowNumber
   * @param pColumnNumber
   * @param pValue
   */
  public void set(
      final int pRowNumber ,
      final int pColumnNumber ,
      final Object pValue )
  {
    this.rowList.get( pRowNumber )[pColumnNumber] = pValue;
  }

  /**
   * add a empty new row.
   * 
   * @return new row count
   */
  public int add()
  {
    this.rowList.add( new Object[this.columnNameArr.length] );
    return this.rowList.size();
  }

  /**
   * add a empty new row.
   * 
   * @return new row count
   */
  public int add(
      final Object[] pRowValueArr )
  {
    if (pRowValueArr.length != this.columnNameArr.length)
    {
      throw new IllegalArgumentException( "param array length "
          + pRowValueArr.length + " is not equal to colum count: " + this.columnNameArr.length );
    }
    this.rowList.add( pRowValueArr );
    return this.rowList.size();
  }

  /**
   * Return value at the specified position.
   * 
   * @param pRowNumber
   * @param pColumnNumber
   * @return
   */
  public Object get(
      final int pRowNumber ,
      final int pColumnNumber )
  {
    return this.rowList.get( pRowNumber )[pColumnNumber];
  }

  /**
   * Return the entire row at the specified
   * position.
   * 
   * @param pRowNumber
   * @return
   */
  public Object[] getRow(
      final int pRowNumber )
  {
    return this.rowList.get( pRowNumber );
  }

  /**
   * Get the names of the columns.
   * 
   * Attention: dont change the columns names
   * to avoid side effects.
   * 
   * @return
   */
  public String[] getColumnNameArr()
  {
    return this.columnNameArr;
  }

  /* (non-Javadoc)
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(
      final Object pObj )
  {
    if (pObj == null)
    {
      return false;
    }
    if ( pObj == this )
    {
      return true;
    }
    if ( !( pObj instanceof Table ))
    {
//System.out.println(1);
      return false;
    }
    final Table pTable = (Table) pObj;
    if ( !Arrays.equals( this.columnNameArr , pTable.columnNameArr ))
    {
      //System.out.println(1);
      return false;
    }

    if (this.rowList.size() != pTable.rowList.size())
    {
//System.out.println(2);
      return false;
    }
    for (int i = 0, n = this.rowList.size(); i < n; i++ )
    {
      final Object[] thisRowArr = this.rowList.get( i );
      final Object[] otherRowArr = pTable.rowList.get( i );
//      if (thisValue == null
//          && otherValue == null)
//      {
//        continue;
//      }
//      if (thisValue != null)
//      {
//        if ( !thisValue.equals( otherValue ))
//        {
//          return false;
//        }
//      }
//      else
//      {
//        if ( !otherValue.equals( thisValue ))
//        {
//          return false;
//        }
//      }
      if ( ! Arrays.equals( thisRowArr, otherRowArr ))
      {
System.out.println( i );      
        return false;
      }
    }

    return true;
  }

  /**
   * @param args
   */
  public static void main(
      String[] args )
  {
  }

}
